#include <PalmOS.h>

#include "resource.h" /* resource IDs first */
#include "open.h"

static FormType *form;
static TableType *table;
static ScrollBarType *scroll;

/* these are used when drawing cells on the fly at run time */
static Int16 screenrows, tabwidth, tabheight, lineheight;

/* and these store the day of week and time of day (in minutes) */
Int16 dayofweek;
opentimetype minuteofday;

/* write a string in a rectangle, cropping bits of it as required */
void WriteInRectangle(Char *s, RectangleType *rect)
{
  Int16 len = StrLen(s);

  if (FntCharsWidth(s, len) <= rect->extent.x) {
    WinDrawChars(s, len, rect->topLeft.x, rect->topLeft.y);
  }
  else {
    /* figure out what we can fit */
    Char *ellipsis = "...";
    Int16 ellwidth = FntCharsWidth(ellipsis, StrLen(ellipsis));
    Int16 partwidth = 0;
    UInt16 i;
    for (i = len - 1; i > 0; --i) {
      if ((partwidth = FntCharsWidth(s, i)) + ellwidth <= rect->extent.x) {
	break; /* draw it this length */
      }
    }
    WinDrawChars(s, i, rect->topLeft.x, rect->topLeft.y);
    WinDrawChars(ellipsis, StrLen(ellipsis), rect->topLeft.x+partwidth, rect->topLeft.y);
  }
}

/* draw the status indicator in a rectangle */
void DrawStatus(Boolean status, RectangleType *rect)
{
  MemHandle bm;
  BitmapType *bmp;

  bm = DmGetResource((DmResType)bitmapRsc, (status?OPENBITMAP:CLOSEDBITMAP));
  bmp = MemHandleLock(bm);
  WinDrawBitmap(bmp,
		rect->topLeft.x + (rect->extent.x - bmp->width) / 2,
		rect->topLeft.y + (rect->extent.y - bmp->height) / 2);
  MemHandleUnlock(bm);
  DmReleaseResource(bm);
}

/* display the time until the next change */
void DrawTimeChange(opentimetype when, RectangleType *rect)
{
  Char tmp[10]; /* this needs to be big enough for the format strings
                   used below */

  if (when >= NEVERCHANGE) {
    StrCopy(tmp, "- --:--");
  }
  else {
    opentimetype d, h, m;
    d = when / MINUTESINDAY;
    h = (when - d*MINUTESINDAY) / MINUTESINHOUR;
    m = (when - d*MINUTESINDAY) % MINUTESINHOUR;
    
    if (d > 0) {
      StrPrintF(tmp, "%d %02d:%02d", d, h, m);
    }
    else {
      StrPrintF(tmp, "%02d:%02d", h, m);
    }
  }
  WinDrawChars(tmp, StrLen(tmp),
	       rect->topLeft.x+(rect->extent.x-FntCharsWidth(tmp, StrLen(tmp))), rect->topLeft.y);
}

/* populate globals with current day of week and time of day in minutes */
void CalcDayAndTime(void)
{
  DateTimeType tmp;

  TimSecondsToDateTime(TimGetSeconds(), &tmp);
  dayofweek = tmp.weekDay;
  minuteofday = tmp.hour * 60 + tmp.minute;
}

/* this draws the table cells */
void MainDrawCell (void *table, Int16 row, Int16 column, RectangleType *bounds)
{
  RectangleType name, status, time; /* we divide the column into three parts */
  MemHandle rec = 0; /* the record */
  UInt16 index; /* index of the record */
  /* next two are the two parts of the record */
  openclose oc; /* the times */
  Char s[MAXNAMELENGTH+1]; /* and the name string */
  Boolean isopen; /* whether this one is currently open */
  opentimetype willchange; /* how long until the open-ness changes */

  /* compute the size of each rectangle */
  /* name is half the width */
  name.topLeft = bounds->topLeft;
  name.extent.x = bounds->extent.x * 23 / 32;
  name.extent.y = bounds->extent.y;
  /* status is a square the height of the cell */
  status.topLeft.x = name.topLeft.x + name.extent.x;
  status.topLeft.y = name.topLeft.y;
  status.extent.x = status.extent.y = bounds->extent.y;
  /* time is the rest */
  time.topLeft.x = status.topLeft.x + status.extent.x;
  time.topLeft.y = status.topLeft.y;
  time.extent.x = bounds->extent.x - (name.extent.x + status.extent.x);
  time.extent.y = bounds->extent.y;
  /* fetch the correct record */
  index = prefs.firstidx;
  if (FindIndexFromRow(&db, row, &index) != errNone || !(rec = QueryRecord(&db, index, &oc, s))) {
    /* display it as an error */
    StrCopy(s, "<error>");
  }
  if (rec) {
    CheckIfOpen(&oc, &isopen, &willchange);
  }
  else {
    isopen = false;
    willchange = 0;
  }
  /* now draw all of that stuff */
  WinEraseRectangle(bounds, 0);
  WriteInRectangle(s, &name);
  DrawStatus(isopen, &status);
  DrawTimeChange(willchange, &time);
}

/* initialize all the form elements, callbacks, etc. */
void MainInitForm()
{
  Int16 maxrows;
  RectangleType r;
  Int16 row;

  /* these will all be useful */
  form = FrmGetActiveForm();
  table = GetObjectPtr(form, MainTable);
  scroll = GetObjectPtr(form, MainScroll);
  /* compute the bounds of the table, number of displayable rows, etc. */
  maxrows = TblGetNumberOfRows(table);
  TblGetBounds(table, &r);
  tabwidth = r.extent.x;
  tabheight = r.extent.y;
  lineheight = FntLineHeight();
  screenrows = tabheight / lineheight;
  for (row = 0; row < maxrows; ++row) {
    TblSetRowHeight(table, row, lineheight);
    TblSetItemStyle(table, row, 0, customTableItem);
    TblSetRowSelectable(table, row, true);
    TblSetRowUsable(table, row, false);
  }
  TblSetCustomDrawProcedure(table, 0, (TableDrawItemFuncPtr) MainDrawCell);
  TblSetColumnUsable(table, 0, true);
  TblHasScrollBar(table, true);
}

void MainUpdateData()
{
  Int16 numlisted;
  Int16 row;
  static Char tmp[25]; /* needs to be large enough for "Day HH:MM am" */
  MemHandle daystr;
  Char *daystrptr;

  numlisted = DmNumRecordsInCategory(db, prefs.category);
  /* Work out which record should be at the top.  It can't be less
     than zero of course, but also we don't allow whitespace at the
     bottom if there are enough rows to fill the display.  However
     we leave the display as close as possible to where the user put
     it last. [Copied from Keyring] */
  if (prefs.firstidx > numlisted - screenrows) {
    prefs.firstidx = numlisted - screenrows;
  }
  if (prefs.firstidx < 0) {
    prefs.firstidx = 0;
  }
  /* make those rows that have valid data visible, and all others invisible */
  for (row = 0; row < screenrows; ++row) {
    TblSetRowUsable(table, row, (row + prefs.firstidx < numlisted) ? true : false);
  }
  if (numlisted > screenrows) {
    SclSetScrollBar(GetObjectPtr(form, MainScroll), (Int16)prefs.firstidx, 0, numlisted - screenrows, screenrows);
  }
  else {
    SclSetScrollBar(GetObjectPtr(form, MainScroll), 0, 0, 0, 0);
  }
  /* update the clock */
  CalcDayAndTime();
  daystr = DmGetResource((DmResType)strRsc, SundayStr + dayofweek);
  daystrptr = MemHandleLock(daystr);
  if (prefs.hour24) {
    StrPrintF(tmp, "%s %02d:%02d", daystrptr,
	      minuteofday / MINUTESINHOUR,
	      minuteofday % MINUTESINHOUR);
  }
  else {
    opentimetype hour = minuteofday / MINUTESINHOUR;
    Boolean pm = hour >= 12;
    hour = ((hour == 0) ? 12 : ((hour > 12) ? hour-12 : hour));
    StrPrintF(tmp, "%s %d:%02d %s", daystrptr,
	      hour, minuteofday % MINUTESINHOUR,
	      pm ? "pm" : "am");
  }
  MemHandleUnlock(daystr);
  DmReleaseResource(daystr);
  FldSetTextPtr(GetObjectPtr(form, MainClock), tmp);
}

void MainUpdateDisplay()
{
  MainUpdateData();
  // FrmDrawForm(form);
  TblDrawTable(table);
  FldDrawField(GetObjectPtr(form, MainClock));
}

Boolean MainFormHandler(EventType *e)
{
  static Char curcatname[dmCategoryLength]; /* static because it's
                                               passed to the trigger
                                               control as it's label */
  UInt16 i;

  switch (e->eType) {

  case frmOpenEvent:
    CalcDayAndTime();
    SortRecords(&db);
    MainInitForm();
    CategoryGetName(db, prefs.category, curcatname);
    CtlSetLabel(GetObjectPtr(form, MainCategoryTrigger), curcatname);
    MainUpdateData(); /* stopped doing a MainUpdate() here - the
                          screen flickers */
    FrmDrawForm(form);
    return false;

  case frmCloseEvent:
    FrmEraseForm(form);
    FrmDeleteForm(form);
    return true;

  case ctlSelectEvent:
    switch (e->data.ctlSelect.controlID) {
      
    case MainNew:
      SetNewRecord();
      FrmGotoForm(PlaceForm);
      return true;

    case MainCategoryTrigger:
      (void) CategorySelect(db, form,
			    MainCategoryTrigger, MainCategoryList,
			    true /* in window header */,
			    &(prefs.category), curcatname,
			    1, EditCatStr);
      CtlSetLabel(GetObjectPtr(form, MainCategoryTrigger), curcatname);
      MainUpdateDisplay();
      return true;

    case MainSortName:
    case MainSortRemaining:
      prefs.sortdir = e->data.ctlSelect.controlID;
      CalcDayAndTime();
      SortRecords(&db);
      MainUpdateDisplay();
      return true;

    default:
      return false;
    }
    return false;

  case sclRepeatEvent:

    switch (e->data.sclRepeat.scrollBarID) {

    case MainScroll:
      prefs.firstidx = e->data.sclRepeat.newValue;
      MainUpdateDisplay();
      return false;

    default:
      return false;
    }
    return false;

  case tblSelectEvent:

    switch (e->data.tblSelect.tableID) {

    case MainTable:
      i = prefs.firstidx;
      FindIndexFromRow(&db, e->data.tblSelect.row, &i);
      SetRecord(i);
      FrmGotoForm(PlaceForm);
      return true;

    default:
      return false;
    }
    return false;

  case menuEvent:
    switch (e->data.menu.itemID) {

    case MainHelp:
      FrmHelp(MainHelpMenu);
      return true;

    case MainAbout:
      FrmGotoForm(AboutForm);
      return true;

    default:
      return false;
    }
    return false;

  case keyDownEvent:

    switch (e->data.keyDown.chr) {

    case vchrPageDown:
      prefs.firstidx += screenrows;
      MainUpdateDisplay();
      return true;

    case vchrPageUp:
      prefs.firstidx -= screenrows;
      MainUpdateDisplay();
      return true;

    default:
      MainUpdateDisplay();
      return false;
      break;

    }
    return false;

  case nilEvent:
    if ((TimGetSeconds()%60) == 0) {
      MainUpdateDisplay(); /* the main loop times out on second
			      boundaries, so that the display can be
			      updated regularly */
    }
    return true;

  default:
    return false;
  }

  return false;
}
